import os
import shutil
import traceback
import uuid
from pathlib import Path
from typing import Dict, List, Set

from ..common import jsonpickle
from ..config import config
from ..utils import (artifact_classes, assemble_path_object,
                     catalog_of_local_artifact, convert_dflow_list, copy_file,
                     expand, flatten, randstr, remove_empty_dir_tag)
from .opio import Artifact, BigParameter, NestedDict, Parameter


def get_slices(path_object, slices):
    if isinstance(slices, int):
        return path_object[slices]
    elif isinstance(slices, str):
        tmp = path_object
        for f in slices.split("."):
            if isinstance(tmp, dict):
                tmp = tmp[f]
            elif isinstance(tmp, list):
                tmp = tmp[int(f)]
        return tmp
    elif isinstance(slices, list):
        return [get_slices(path_object, i) for i in slices]
    return path_object


def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
                          sub_path=None, n_parts=None, keys_of_parts=None,
                          path=None, prefix=None):
    root = None
    if n_parts is not None:
        path_object = []
        for i in range(n_parts):
            art_path = '%s/inputs/artifacts/dflow_%s_%s' % (data_root, name, i)
            if config["detect_empty_dir"]:
                remove_empty_dir_tag(art_path)
            po = assemble_path_object(art_path)
            if prefix is not None:
                po = get_slices(po, prefix[i])
            if not po:
                path_object.append(art_path)
            elif isinstance(po, list):
                path_object += po
            else:
                path_object.append(po)
    elif keys_of_parts is not None:
        path_object = {}
        for i in keys_of_parts:
            art_path = '%s/inputs/artifacts/dflow_%s_%s' % (data_root, name, i)
            if config["detect_empty_dir"]:
                remove_empty_dir_tag(art_path)
            po = assemble_path_object(art_path)
            if prefix is not None:
                po = get_slices(po, prefix[i])
            if not po:
                path_object[i] = art_path
            else:
                path_object[i] = po
        path_object = expand(path_object)
    else:
        art_path = '%s/inputs/artifacts/%s' % (data_root, name) \
            if path is None else path
        if sub_path is not None:
            art_path = os.path.join(art_path, sub_path)
        root = art_path
        if not os.path.exists(art_path):  # for optional artifact
            return None
        if config["detect_empty_dir"]:
            remove_empty_dir_tag(art_path)
        path_object = assemble_path_object(art_path)
        path_object = get_slices(path_object, prefix)

    path_object = get_slices(path_object, slices)

    if sign.type in [str, Path]:
        if path_object is None or isinstance(path_object, str):
            res = path_object
        elif isinstance(path_object, list) and len(path_object) == 1 and (
            path_object[0] is None or isinstance(path_object[0], str)) \
                and sign.sub_path:
            res = path_object[0]
        else:
            res = art_path
        res = path_or_none(res) if sign.type == Path else res
    elif sign.type in [List[str], List[Path], Set[str], Set[Path]]:
        if path_object is None:
            return None
        elif isinstance(path_object, str):
            res = [path_object]
        elif isinstance(path_object, list) and all([
                p is None or isinstance(p, str) for p in path_object]):
            res = path_object
        else:
            res = list(flatten(path_object).values())

        if sign.type == List[str]:
            pass
        elif sign.type == List[Path]:
            res = path_or_none(res)
        elif sign.type == Set[str]:
            res = set(res)
        else:
            res = set(path_or_none(res))
    elif sign.type in [Dict[str, str], NestedDict[str]]:
        res = path_object
    elif sign.type in [Dict[str, Path], NestedDict[Path]]:
        res = path_or_none(path_object)

    if res is None:
        return None

    _cls = res.__class__
    res = artifact_classes[_cls](res)
    res.art_root = root
    return res


def path_or_none(p):
    if p is None:
        return None
    elif isinstance(p, list):
        return [path_or_none(i) for i in p]
    elif isinstance(p, dict):
        return {k: path_or_none(v) for k, v in p.items()}
    else:
        return Path(p)


def handle_input_parameter(name, value, sign, slices=None, data_root="/tmp"):
    if "dflow_list_item" in value:
        dflow_list = []
        for item in jsonpickle.loads(value):
            dflow_list += jsonpickle.loads(item)
        obj = convert_dflow_list(dflow_list)
    elif isinstance(sign, BigParameter) and config["mode"] != "debug":
        with open(data_root + "/inputs/parameters/" + name, "r") as f:
            content = jsonpickle.loads(f.read())
            obj = content
    else:
        if isinstance(sign, (Parameter, BigParameter)):
            sign = sign.type
        if sign == str and slices is None:
            obj = value
        else:
            obj = jsonpickle.loads(value)

    if slices is not None:
        assert isinstance(
            obj, list), "Only parameters of type list can be sliced, while %s"\
            " is not list" % obj
        if isinstance(slices, list):
            obj = [obj[i] for i in slices]
        else:
            obj = obj[slices]

    return obj


def slice_to_dir(slice):
    return str(slice).replace(".", "/")


def handle_output_artifact(name, value, sign, slices=None, data_root="/tmp",
                           create_dir=False):
    path_list = []
    if sign.type in [str, Path]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        if slices is None:
            slices = 0
        path_list.append(copy_results_and_return_path_item(
            value, name, slices, data_root,
            slice_to_dir(slices) if create_dir else None))
    elif sign.type in [List[str], List[Path], Set[str], Set[Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        if slices is not None:
            if isinstance(slices, int):
                for path in value:
                    path_list.append(copy_results_and_return_path_item(
                        path, name, slices, data_root,
                        slice_to_dir(slices) if create_dir else None))
            else:
                assert len(slices) == len(value)
                for path, s in zip(value, slices):
                    if isinstance(path, list):
                        for p in path:
                            path_list.append(
                                copy_results_and_return_path_item(
                                    p, name, s, data_root, slice_to_dir(
                                        s) if create_dir else None))
                    else:
                        path_list.append(copy_results_and_return_path_item(
                            path, name, s, data_root, slice_to_dir(
                                s) if create_dir else None))
        else:
            for s, path in enumerate(value):
                path_list.append(copy_results_and_return_path_item(
                    path, name, s, data_root))
    elif sign.type in [Dict[str, str], Dict[str, Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        for s, path in value.items():
            path_list.append(copy_results_and_return_path_item(
                path, name, s, data_root))
    elif sign.type in [NestedDict[str], NestedDict[Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        for s, path in flatten(value).items():
            path_list.append(copy_results_and_return_path_item(
                path, name, s, data_root))

    os.makedirs(data_root + "/outputs/artifacts/%s/%s" % (name, config[
        "catalog_dir_name"]), exist_ok=True)
    with open(data_root + "/outputs/artifacts/%s/%s/%s" % (name, config[
            "catalog_dir_name"], uuid.uuid4()), "w") as f:
        f.write(jsonpickle.dumps({"path_list": path_list}))
    if config["detect_empty_dir"]:
        handle_empty_dir(data_root + "/outputs/artifacts/%s" % name)
    if config["save_path_as_parameter"]:
        with open(data_root + '/outputs/parameters/dflow_%s_path_list'
                  % name, 'w') as f:
            f.write(jsonpickle.dumps(path_list))


def handle_output_parameter(name, value, sign, slices=None, data_root="/tmp"):
    if slices is not None:
        if isinstance(slices, list):
            assert isinstance(value, list) and len(slices) == len(value)
            res = [{"dflow_list_item": v, "order": s}
                   for v, s in zip(value, slices)]
        else:
            res = [{"dflow_list_item": value, "order": slices}]
        with open(data_root + '/outputs/parameters/' + name, 'w') as f:
            f.write(jsonpickle.dumps(res))
    elif isinstance(sign, BigParameter) and config["mode"] != "debug":
        with open(data_root + "/outputs/parameters/" + name, "w") as f:
            f.write(jsonpickle.dumps(value))
    else:
        if isinstance(sign, (Parameter, BigParameter)):
            sign = sign.type
        if sign == str:
            with open(data_root + '/outputs/parameters/' + name, 'w') as f:
                f.write(value)
        else:
            with open(data_root + '/outputs/parameters/' + name, 'w') as f:
                f.write(jsonpickle.dumps(value))


def copy_results_and_return_path_item(path, name, order, data_root="/tmp",
                                      slice_dir=None):
    if path and os.path.exists(str(path)):
        return {"dflow_list_item": copy_results(
                    path, name, data_root, slice_dir), "order": order}
    else:
        return {"dflow_list_item": None, "order": order}


def copy_results(source, name, data_root="/tmp", slice_dir=None):
    source = str(source)
    # if refer to input artifact
    if source.find(data_root + "/inputs/artifacts/") == 0:
        # retain original directory structure
        i = source.find("/", len(data_root + "/inputs/artifacts/"))
        if i == -1:
            rel_path = randstr()
        else:
            rel_path = source[i+1:]
        if slice_dir is not None:
            rel_path = "%s/%s" % (slice_dir, rel_path)
        target = data_root + "/outputs/artifacts/%s/%s" % (name, rel_path)
        copy_file(source, target, shutil.copy)
        if rel_path[:1] == "/":
            rel_path = rel_path[1:]
        return rel_path
    else:
        cwd = os.getcwd()
        if not cwd.endswith("/"):
            cwd = cwd + "/"
        if source.startswith(cwd):
            source = source[len(cwd):]
        rel_path = source[1:] if source[:1] == "/" else source
        if slice_dir is not None:
            rel_path = "%s/%s" % (slice_dir, rel_path)
        target = data_root + "/outputs/artifacts/%s/%s" % (name, rel_path)
        copy_file(source, target)
        return rel_path


def handle_empty_dir(path):
    # touch an empty file in each empty dir, as object storage will ignore
    # empty dirs
    for dn, ds, fs in os.walk(path, followlinks=True):
        if len(ds) == 0 and len(fs) == 0:
            with open(os.path.join(dn, ".empty_dir"), "w"):
                pass


def handle_lineage(wf_name, pod_name, op_obj, input_urns, workflow_urn,
                   data_root="/tmp"):
    task_name = wf_name + "/" + pod_name
    output_sign = op_obj.get_output_sign()
    output_uris = {}
    for name, sign in output_sign.items():
        if isinstance(sign, Artifact):
            output_uris[name] = op_obj.get_output_artifact_storage_key(name)
    input_urns = {name: jsonpickle.loads(urn) if urn[:1] == "["
                  else urn for name, urn in input_urns.items()}
    output_urns = config["lineage"].register_task(
        task_name, input_urns, output_uris, workflow_urn)
    for name, urn in output_urns.items():
        with open("%s/outputs/parameters/dflow_%s_urn" % (data_root, name),
                  "w") as f:
            f.write(urn)


def absolutize(path):
    if path is None:
        return None
    if isinstance(path, str):
        return os.path.abspath(path)
    if isinstance(path, Path):
        return path.absolute()
    if isinstance(path, list):
        return [absolutize(p) for p in path]
    if isinstance(path, dict):
        return {k: absolutize(p) for k, p in path.items()}


def try_to_execute(input, slice_dir, op_obj, output_sign, cwd):
    os.chdir(cwd)
    if slice_dir is not None:
        os.makedirs(slice_dir, exist_ok=True)
        os.chdir(slice_dir)
    try:
        output = op_obj.execute(input)
        for n, s in output_sign.items():
            if isinstance(s, Artifact):
                output[n] = absolutize(output[n])
        os.chdir(cwd)
        return output, None
    except Exception as e:
        traceback.print_exc()
        os.chdir(cwd)
        return None, e


def get_input_slices(name, data_root="/tmp"):
    art_path = '%s/inputs/artifacts/%s' % (data_root, name)
    catalog = catalog_of_local_artifact(art_path)
    slices = [item["order"] for item in catalog]
    slices.sort()
    return slices



from copy import deepcopy
from typing import Union, List, Optional, Any

from dflow.executor import Executor
from dflow.op_template import ScriptOPTemplate
from dflow.utils import randstr

try:
    from argo.workflows.client import (V1Volume, V1VolumeMount,
                                       V1alpha1UserContainer,
                                       V1EmptyDirVolumeSource)
except ImportError:
    pass


def _ray_init_container(main_image: str,
                        install_mirror: Union[bool, str, None] = None,
                        ray_remote_path_prefix: str = '/tmp/ray',
                        container_name: str = 'ray-init',
                        ):
    """Get init_container with the same image as main_image do. If the
    container have python with ray (import ray), skip the install; else run pip
    install ray.

    Args:
        main_image: image name of main container
        install_mirror: mirror url used by pip install,
                such as tuna: "https://pypi.tuna.tsinghua.edu.cn/simple"
        ray_remote_path_prefix: ray install path (default: "/tmp/ray")

    Returns:
    """
    RAY_INSTALL_STATEMENT = 'pip install ray ' \
                            f'--target={ray_remote_path_prefix}'
    if install_mirror:
        RAY_INSTALL_STATEMENT += f' -i {install_mirror}'
    return V1alpha1UserContainer(
        name=container_name,
        image=main_image,
        command=['sh'],
        args=['-c',
              f"python -c 'import ray' >/dev/null 2>&1 ||"
              f' {{ {RAY_INSTALL_STATEMENT}; }}'],
        volume_mounts=[V1VolumeMount(
            name='ray-python-packages', mount_path=ray_remote_path_prefix)],
    )


class RayClusterExecutor(Executor):

    def __init__(
            self,
            ray_host: str,
            ray_remote_path: str = '/tmp/ray',
            workdir: str = '~/dflow/workflows/{{workflow.name}}/{{pod.name}}',
            ray_install_mirror=None,
            ray_dependencies: Optional[List[Any]] = None,
    ) -> None:
        """Ray cluster executor.

        Args:
            ray_host:
            ray_remote_path:
            workdir:
            ray_install_mirror:
            ray_dependencies: `py_modules` of ray.init(runtime_env={})
        """
        if ray_dependencies is None:
            ray_dependencies = []
        self.ray_host = ray_host
        self.ray_remote_path = ray_remote_path
        self.workdir = workdir
        self.ray_install_mirror = ray_install_mirror
        self.ray_dependencies = ray_dependencies

    def render(self, template: ScriptOPTemplate):
        new_template = deepcopy(template)
        new_template.name += '-' + randstr()
        if new_template.init_containers is None:
            new_template.init_containers = [
                _ray_init_container(template.image, self.ray_install_mirror,
                                    self.ray_remote_path)]
        else:
            assert isinstance(new_template.init_containers,
                              list), f'Type of current init_containers is ' \
                                     f'not list, but ' \
                                     f'{type(new_template.init_containers)}, '\
                                     f'check your other ' \
                                     f'init_container settings.'
            new_template.init_containers.append(
                _ray_init_container(template.image, self.ray_install_mirror,
                                    self.ray_remote_path))
        new_template.script = 'import sys,os                                ' \
                              '\n' \
                              f"sys.path.append('{self.ray_remote_path}')   " \
                              f"\n" \
                              "if os.environ.get('RAY_ADDRESS') is None:    " \
                              "\n" \
                              f"    os.environ['RAY_ADDRESS']=" \
                              f"'{self.ray_host}' " \
                              f"\n" \
                              'else:                                        ' \
                              '\n' \
                              "    print(f\"Not use input ray_host address," \
                              " use {os.environ['RAY_ADDRESS']} instead.\") " \
                              "\n" \
                              + template.script
        # To locate the initialization of package path in `python_op_template`.
        insert_index = new_template.script.find('from dflow import config')
        new_script = list(new_template.script)
        _dependencies_str = ','.join(item.__name__
                                     for item in self.ray_dependencies)
        ray_dependencies_import = \
            f"import {_dependencies_str}\n" if len(
                self.ray_dependencies) > 0 else '\n'
        new_script.insert(
            insert_index,
            'import ray\n' +
            ray_dependencies_import +
            "ray.init(os.environ['RAY_ADDRESS'], runtime_env={"
            f"'py_modules':["
            f"{','.join(item.__name__ for item in self.ray_dependencies)}"
            ']})\n')
        new_template.script = ''.join(new_script)
        new_template.volumes.append(V1Volume(
            name='ray-python-packages', empty_dir=V1EmptyDirVolumeSource()))
        new_template.mounts.append(V1VolumeMount(
            name='ray-python-packages', mount_path=self.ray_remote_path))
        return new_template



import logging
import os
import time
from copy import deepcopy
from typing import Dict, List, Optional, Union

from .common import (input_artifact_pattern, input_parameter_pattern,
                     step_output_artifact_pattern,
                     step_output_parameter_pattern)
from .config import config, s3_config
from .context_syntax import GLOBAL_CONTEXT
from .io import Inputs, Outputs
from .op_template import OPTemplate
from .step import Step, add_slices
from .utils import ProcessPoolExecutor

try:
    from argo.workflows.client import V1alpha1Metadata, V1alpha1Template
except Exception:
    pass


class Steps(OPTemplate):
    """
    Steps

    Args:
        name: the name of the steps
        inputs: inputs in the template
        outputs: outputs in the template
        steps: a sequential list of steps
        memoize_key: memoized key of the steps
        annotations: annotations for the OP template
        parallelism: maximum number of running pods for the OP template
        """

    def __init__(
            self,
            name: Optional[str] = None,
            inputs: Optional[Inputs] = None,
            outputs: Optional[Outputs] = None,
            steps: List[Union[Step, List[Step]]] = None,
            memoize_key: Optional[str] = None,
            annotations: Dict[str, str] = None,
            parallelism: Optional[int] = None,
    ) -> None:
        super().__init__(name=name, inputs=inputs, outputs=outputs,
                         memoize_key=memoize_key, annotations=annotations)
        self.parallelism = parallelism
        self.steps = []
        if steps is not None:
            for step in steps:
                self.add(step)

    @classmethod
    def from_dict(cls, d, templates):
        kwargs = {
            "name": d.get("name", None),
            "inputs": Inputs.from_dict(d.get("inputs", {})),
            "outputs": Outputs.from_dict(d.get("outputs", {})),
            "memoize_key": d.get("memoize", {}).get("key", None),
            "annotations": d.get("metadata", {}).get("annotations", None),
            "parallelism": d.get("parallelism", None),
        }
        kwargs["steps"] = [[Step.from_dict(ps, templates) for ps in step]
                           for step in d.get("steps", [[]])]
        return cls(**kwargs)

    def __iter__(self):
        return iter(self.steps)

    def add(
            self,
            step: Union[Step, List[Step]],
    ) -> None:
        """
        Add a step or a list of parallel steps to the steps

        Args:
            step: a step or a list of parallel steps to be added to the
                entrypoint of the workflow
        """
        assert isinstance(step, (Step, list))
        if isinstance(step, Step):
            if step.prepare_step is not None:
                self.steps.append(step.prepare_step)
        elif isinstance(step, list):
            prepare_steps = [ps.prepare_step for ps in step
                             if ps.prepare_step is not None]
            if prepare_steps:
                self.steps.append(prepare_steps)
        for ps in [step] if isinstance(step, Step) else step:
            if ps.parallelism is not None:
                if self.parallelism is None:
                    self.parallelism = ps.parallelism
                else:
                    self.parallelism = min(self.parallelism, ps.parallelism)
                    logging.warn("Multiple steps specified parallelism, small"
                                 "er one %s will be used" % self.parallelism)
        self.steps.append(step)
        if isinstance(step, Step):
            if step.check_step is not None:
                self.steps.append(step.check_step)
        elif isinstance(step, list):
            check_steps = [ps.check_step for ps in step
                           if ps.check_step is not None]
            if check_steps:
                self.steps.append(check_steps)

    def convert_to_argo(self, memoize_prefix=None,
                        memoize_configmap="dflow", context=None):
        argo_steps = []
        templates = []
        assert len(self.steps) > 0, "Steps %s is empty" % self.name
        for step in self.steps:
            # each step of steps should be a list of parallel steps, if not,
            # create a sigleton
            if not isinstance(step, list):
                step = [step]
            argo_parallel_steps = []
            for ps in step:
                argo_parallel_steps.append(ps.convert_to_argo(context))
                # template may change after conversion
                templates.append(ps.template)
            argo_steps.append(argo_parallel_steps)

        self.handle_key(memoize_prefix, memoize_configmap)
        argo_template = \
            V1alpha1Template(name=self.name,
                             metadata=V1alpha1Metadata(
                                 annotations=self.annotations),
                             steps=argo_steps,
                             inputs=self.inputs.convert_to_argo(),
                             outputs=self.outputs.convert_to_argo(),
                             memoize=self.memoize,
                             parallelism=self.parallelism)
        return argo_template, templates

    def convert_to_graph(self):
        graph_steps = []
        templates = []
        for step in self.steps:
            if not isinstance(step, list):
                step = [step]
            graph_parallel_steps = []
            for ps in step:
                if not ps.name.endswith("-init-artifact") and \
                    not ps.name.endswith("-check-num-success") and \
                        not ps.name.endswith("-check-success-ratio"):
                    graph_parallel_steps.append(ps.convert_to_graph())
                    templates.append(ps.template)
            if len(graph_parallel_steps) > 0:
                graph_steps.append(graph_parallel_steps)

        return {
            "type": "Steps",
            "name": self.name,
            "annotations": self.annotations,
            "steps": graph_steps,
            "inputs": self.inputs.convert_to_graph(),
            "outputs": self.outputs.convert_to_graph(),
            "parallelism": self.parallelism,
        }, templates

    @classmethod
    def from_graph(cls, graph, templates):
        assert graph.pop("type") == "Steps"
        graph["inputs"] = Inputs.from_graph(graph.get("inputs", {}))
        steps = graph.pop("steps")
        outputs = graph.pop("outputs", {})

        obj = cls(**graph)
        templates[graph["name"]] = obj

        step_dict = {}
        for step in steps:
            parallel_steps = []
            for ps in step:
                s = Step.from_graph(ps, templates)
                parallel_steps.append(s)
                step_dict[s.name] = s
            obj.add(parallel_steps)

        # replace variable references with pointers
        for step in obj.steps:
            for ps in step:
                for name, par in ps.inputs.parameters.items():
                    value = getattr(par, "value", None)
                    if isinstance(value, str):
                        match = input_parameter_pattern.match(value)
                        if match:
                            ps.set_parameters({
                                name: obj.inputs.parameters[match.group(1)]})
                        match = step_output_parameter_pattern.match(value)
                        if match:
                            ps.set_parameters({name: step_dict[match.group(
                                1)].outputs.parameters[match.group(2)]})
                for name, art in ps.inputs.artifacts.items():
                    source = art.source
                    if isinstance(source, str):
                        match = input_artifact_pattern.match(source)
                        if match:
                            ps.set_artifacts({
                                name: obj.inputs.artifacts[match.group(1)]})
                        match = step_output_artifact_pattern.match(source)
                        if match:
                            ps.set_artifacts({name: step_dict[match.group(
                                1)].outputs.artifacts[match.group(2)]})

        obj.outputs = Outputs.from_graph(outputs)
        # replace variable references with pointers
        for par in obj.outputs.parameters.values():
            _from = par.value_from_parameter
            if isinstance(_from, str):
                match = input_parameter_pattern.match(_from)
                if match:
                    par.value_from_parameter = obj.inputs.parameters[
                        match.group(1)]
                match = step_output_parameter_pattern.match(_from)
                if match:
                    par.value_from_parameter = step_dict[match.group(1)].\
                        outputs.parameters[match.group(2)]
        for art in obj.outputs.artifacts.values():
            _from = art._from
            if isinstance(_from, str):
                match = input_artifact_pattern.match(_from)
                if match:
                    art._from = obj.inputs.artifacts[match.group(1)]
                match = step_output_artifact_pattern.match(_from)
                if match:
                    art._from = step_dict[match.group(1)].outputs.\
                        artifacts[match.group(2)]
        return obj

    def run(self, workflow_id=None, context=None):
        self.workflow_id = workflow_id
        for step in self:
            if isinstance(step, list):
                import concurrent.futures
                cwd = os.getcwd()
                max_workers = config["debug_pool_workers"]
                if max_workers == -1:
                    max_workers = len(step)
                if max_workers is None:
                    max_workers = os.cpu_count() or 1
                max_workers = min(max_workers, len(step)) or 1
                with ProcessPoolExecutor(max_workers) as pool:
                    futures = []
                    self_copy = deepcopy(self)
                    for i, ps in enumerate(step):
                        ps.phase = "Pending"
                        try:
                            future = pool.submit(
                                ps.run_with_config, self_copy, context, config,
                                s3_config, cwd)
                        except concurrent.futures.process.BrokenProcessPool \
                                as e:
                            # retrieve exception of subprocess before exit
                            for future in concurrent.futures.as_completed(
                                    futures):
                                future.result()
                            raise e
                        futures.append(future)
                        if config["debug_batch_size"] and i != len(step) - 1 \
                                and (i+1) % config["debug_batch_size"] == 0:
                            logging.info(
                                "Wait %s seconds before submitting next "
                                "batch" % config["debug_batch_interval"])
                            time.sleep(config["debug_batch_interval"])

                    for future in concurrent.futures.as_completed(futures):
                        j = futures.index(future)
                        try:
                            ps = future.result()
                        except Exception:
                            import traceback
                            traceback.print_exc()
                            step[j].phase = "Failed"
                            if not step[j].continue_on_failed:
                                raise RuntimeError("Step %s failed" % step[j])
                        else:
                            step[j].outputs = deepcopy(ps.outputs)
                            logging.info("Outputs of %s collected" % step[j])
            else:
                step.run(self, context)

    def add_slices(self, slices, layer=0):
        add_slices(self, slices, layer=layer)

    def copy(self):
        new_template = self.deepcopy()
        for step, new_step in zip(self.steps, new_template.steps):
            if isinstance(new_step, list):
                for ps, new_ps in zip(step, new_step):
                    new_ps.template = ps.template
            else:
                new_step.template = step.template
        return new_template

    def __enter__(self) -> 'Steps':
        GLOBAL_CONTEXT.in_context = True
        GLOBAL_CONTEXT.current_workflow = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        GLOBAL_CONTEXT.in_context = False



import inspect
import json

from .common import (input_artifact_pattern, input_parameter_pattern,
                     step_output_artifact_pattern,
                     step_output_parameter_pattern,
                     task_output_artifact_pattern,
                     task_output_parameter_pattern)
from .common import jsonpickle
from .config import config
from .dag import DAG
from .io import InputArtifact, InputParameter, OutputArtifact, OutputParameter
from .op_template import ScriptOPTemplate
from .python import PythonOPTemplate
from .step import Step
from .steps import Steps
from .utils import Variable, evalable_repr
from .workflow import Workflow


class CodeGenerator:
    def __init__(self, graph):
        self.graph = graph
        self.imports = set()
        self.code = ""

    def get_var_name(self, name):
        return name.replace("-", "_")

    def get_kwargs(self, template, cls):
        kwargs = {}
        sign = inspect.signature(cls.__init__).parameters
        for k, v in template.items():
            if k in sign:
                if v == sign[k].default:
                    continue
                if sign[k].default is None and v in [[], {}]:
                    continue
            kwargs[k] = v
        return kwargs

    def render_python_op_template(self, var_name, template):
        assert template.pop("type") == "PythonOPTemplate"
        op = template.pop("op")
        if op["module"] in ["__main__", "__mp_main__"]:
            self.code += op["source"]
        else:
            self.imports.add((op["module"], op["name"]))

        kwargs = self.get_kwargs(template, PythonOPTemplate)
        if "command" in kwargs and kwargs["command"] == ["python3"]:
            del kwargs["command"]
        kwargs["op_class"] = Variable(op["name"])
        self.imports.add(("dflow.python", "PythonOPTemplate"))
        self.code += "%s = PythonOPTemplate(%s)\n" % (var_name, ", ".join([
            "%s=%s" % (k, evalable_repr(v, self.imports))
            for k, v in kwargs.items()]))

    def render_script_op_template(self, var_name, template):
        assert template.pop("type") == "ScriptOPTemplate"
        kwargs = self.get_kwargs(template, ScriptOPTemplate)
        self.imports.add(("dflow", "ScriptOPTemplate"))
        self.code += "%s = ScriptOPTemplate(%s)\n" % (var_name, ", ".join([
            "%s=%s" % (k, evalable_repr(v, self.imports))
            for k, v in kwargs.items()]))

    def render_steps(self, var_name, template):
        assert template.pop("type") == "Steps"
        inputs = template.pop("inputs", {})
        outputs = template.pop("outputs", {})
        steps = template.pop("steps", {})
        kwargs = self.get_kwargs(template, Steps)
        self.imports.add(("dflow", "Steps"))
        self.code += "%s = Steps(%s)\n" % (var_name, ", ".join([
            "%s=%s" % (k, evalable_repr(v, self.imports))
            for k, v in kwargs.items()]))

        for name, par in inputs.get("parameters", {}).items():
            kwargs = self.get_kwargs(par, InputParameter)
            kwargs.pop("name", None)
            self.imports.add(("dflow", "InputParameter"))
            self.code += "%s.inputs.parameters['%s'] = InputParameter(%s)\n" \
                % (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        for name, art in inputs.get("artifacts", {}).items():
            kwargs = self.get_kwargs(art, InputArtifact)
            kwargs.pop("name", None)
            self.imports.add(("dflow", "InputArtifact"))
            self.code += "%s.inputs.artifacts['%s'] = InputArtifact(%s)\n" % \
                (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        step_dict = {}
        for step in steps:
            parallel_steps = []
            for ps in step:
                kwargs = self.get_kwargs(ps, Step)
                if "util_image" in kwargs and kwargs["util_image"] == config[
                        "util_image"]:
                    del kwargs["util_image"]
                if kwargs["template"] in self.templates:
                    kwargs["template"] = Variable(
                        self.templates[kwargs["template"]])
                elif kwargs["template"] == template["name"]:
                    kwargs["template"] = Variable(var_name)
                for name, par in kwargs.get("parameters", {}).items():
                    if isinstance(par, str):
                        match = input_parameter_pattern.match(par)
                        if match:
                            kwargs["parameters"][name] = Variable(
                                "%s.inputs.parameters['%s']" % (
                                    var_name, match.group(1)))
                        match = step_output_parameter_pattern.match(par)
                        if match:
                            kwargs["parameters"][name] = Variable(
                                "%s.outputs.parameters['%s']" % (
                                    step_dict[match.group(1)], match.group(2)))
                    elif isinstance(par, dict) and "py/object" in par:
                        self.imports.add(("dflow", "jsonpickle"))
                        kwargs["parameters"][name] = Variable(
                            "jsonpickle.loads(%s)" % repr(json.dumps(par)))
                for name, art in kwargs.get("artifacts", {}).items():
                    if isinstance(art, str):
                        match = input_artifact_pattern.match(art)
                        if match:
                            kwargs["artifacts"][name] = Variable(
                                "%s.inputs.artifacts['%s']" % (
                                    var_name, match.group(1)))
                        match = step_output_artifact_pattern.match(art)
                        if match:
                            kwargs["artifacts"][name] = Variable(
                                "%s.outputs.artifacts['%s']" % (
                                    step_dict[match.group(1)], match.group(2)))

                self.imports.add(("dflow", "Step"))
                step_var_name = "step_" + self.get_var_name(ps["name"])
                self.code += "%s = Step(%s)\n" % (
                    step_var_name, ", ".join(["%s=%s" % (k, evalable_repr(
                        v, self.imports)) for k, v in kwargs.items()]))
                step_dict[ps["name"]] = step_var_name
                parallel_steps.append(step_var_name)

            if len(parallel_steps) == 1:
                self.code += "%s.add(%s)\n" % (var_name, parallel_steps[0])
            else:
                self.code += "%s.add([%s])\n" % (
                    var_name, ", ".join(parallel_steps))

        for name, par in outputs.get("parameters", {}).items():
            kwargs = self.get_kwargs(par, OutputParameter)
            kwargs.pop("name", None)
            _from = kwargs.get("value_from_parameter")
            if isinstance(_from, str):
                match = input_parameter_pattern.match(_from)
                if match:
                    kwargs["value_from_parameter"] = Variable(
                        "%s.inputs.parameters['%s']" % (
                            var_name, match.group(1)))
                match = step_output_parameter_pattern.match(_from)
                if match:
                    kwargs["value_from_parameter"] = Variable(
                        "%s.outputs.parameters['%s']" % (
                            step_dict[match.group(1)], match.group(2)))
            self.imports.add(("dflow", "OutputParameter"))
            self.code += "%s.outputs.parameters['%s'] = OutputParameter(%s)\n"\
                % (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        for name, art in outputs.get("artifacts", {}).items():
            kwargs = self.get_kwargs(art, OutputArtifact)
            kwargs.pop("name", None)
            if "archive" in kwargs and kwargs["archive"] == config[
                    "archive_mode"]:
                del kwargs["archive"]
            _from = kwargs.get("_from")
            if isinstance(_from, str):
                match = input_artifact_pattern.match(_from)
                if match:
                    kwargs["_from"] = Variable(
                        "%s.inputs.artifacts['%s']" % (
                            var_name, match.group(1)))
                match = step_output_artifact_pattern.match(_from)
                if match:
                    kwargs["_from"] = Variable(
                        "%s.outputs.artifacts['%s']" % (
                            step_dict[match.group(1)], match.group(2)))
            self.imports.add(("dflow", "OutputArtifact"))
            self.code += "%s.outputs.artifacts['%s'] = OutputArtifact(%s)\n" %\
                (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

    def render_dag(self, var_name, template):
        assert template.pop("type") == "DAG"
        inputs = template.pop("inputs", {})
        outputs = template.pop("outputs", {})
        tasks = template.pop("tasks", {})
        kwargs = self.get_kwargs(template, DAG)
        self.imports.add(("dflow", "DAG"))
        self.code += "%s = DAG(%s)\n" % (var_name, ", ".join([
            "%s=%s" % (k, evalable_repr(v, self.imports))
            for k, v in kwargs.items()]))

        for name, par in inputs.get("parameters", {}).items():
            kwargs = self.get_kwargs(par, InputParameter)
            kwargs.pop("name", None)
            self.imports.add(("dflow", "InputParameter"))
            self.code += "%s.inputs.parameters['%s'] = InputParameter(%s)\n" \
                % (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        for name, art in inputs.get("artifacts", {}).items():
            kwargs = self.get_kwargs(art, InputArtifact)
            kwargs.pop("name", None)
            self.imports.add(("dflow", "InputArtifact"))
            self.code += "%s.inputs.artifacts['%s'] = InputArtifact(%s)\n" % \
                (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        task_dict = {}
        for task in tasks:
            task_var_name = "task_" + self.get_var_name(task["name"])
            task_dict[task["name"]] = task_var_name

        kwargs_dict = {}
        dependencies_dict = {t: [] for t in task_dict}
        for task in tasks:
            dependencies = task.get("dependencies", [])
            kwargs = self.get_kwargs(task, Step)
            if dependencies:
                dependencies_dict[task["name"]].extend(dependencies)
                kwargs["dependencies"] = [Variable(task_dict[t])
                                          for t in dependencies]
            if "util_image" in kwargs and kwargs["util_image"] == config[
                    "util_image"]:
                del kwargs["util_image"]
            if kwargs["template"] in self.templates:
                kwargs["template"] = Variable(
                    self.templates[kwargs["template"]])
            elif kwargs["template"] == template["name"]:
                kwargs["template"] = Variable(var_name)
            for name, par in kwargs.get("parameters", {}).items():
                if isinstance(par, str):
                    match = input_parameter_pattern.match(par)
                    if match:
                        kwargs["parameters"][name] = Variable(
                            "%s.inputs.parameters['%s']" % (
                                var_name, match.group(1)))
                    match = task_output_parameter_pattern.match(par)
                    if match:
                        kwargs["parameters"][name] = Variable(
                            "%s.outputs.parameters['%s']" % (
                                task_dict[match.group(1)], match.group(2)))
                        dependencies_dict[task["name"]].append(match.group(1))
                elif isinstance(par, dict) and "py/object" in par:
                    self.imports.add(("dflow", "jsonpickle"))
                    kwargs["parameters"][name] = Variable(
                        "jsonpickle.loads(%s)" % repr(json.dumps(par)))
            for name, art in kwargs.get("artifacts", {}).items():
                if isinstance(art, str):
                    match = input_artifact_pattern.match(art)
                    if match:
                        kwargs["artifacts"][name] = Variable(
                            "%s.inputs.artifacts['%s']" % (
                                var_name, match.group(1)))
                    match = task_output_artifact_pattern.match(art)
                    if match:
                        kwargs["artifacts"][name] = Variable(
                            "%s.outputs.artifacts['%s']" % (
                                task_dict[match.group(1)], match.group(2)))
                        dependencies_dict[task["name"]].append(match.group(1))
            kwargs_dict[task["name"]] = kwargs

        while len(kwargs_dict) > 0:
            update = False
            for name, kwargs in list(kwargs_dict.items()):
                if all([t not in kwargs_dict
                        for t in dependencies_dict[name]]):
                    self.imports.add(("dflow", "Task"))
                    task_var_name = "task_" + self.get_var_name(name)
                    self.code += "%s = Task(%s)\n" % (
                        task_var_name, ", ".join(["%s=%s" % (k, evalable_repr(
                            v, self.imports)) for k, v in kwargs.items()]))
                    self.code += "%s.add(%s)\n" % (var_name, task_var_name)
                    update = True
                    del kwargs_dict[name]
            assert update, "Failed to resolve tasks: %s" % list(kwargs_dict)

        for name, par in outputs.get("parameters", {}).items():
            kwargs = self.get_kwargs(par, OutputParameter)
            kwargs.pop("name", None)
            _from = kwargs.get("value_from_parameter")
            if isinstance(_from, str):
                match = input_parameter_pattern.match(_from)
                if match:
                    kwargs["value_from_parameter"] = Variable(
                        "%s.inputs.parameters['%s']" % (
                            var_name, match.group(1)))
                match = task_output_parameter_pattern.match(_from)
                if match:
                    kwargs["value_from_parameter"] = Variable(
                        "%s.outputs.parameters['%s']" % (
                            task_dict[match.group(1)], match.group(2)))
            self.imports.add(("dflow", "OutputParameter"))
            self.code += "%s.outputs.parameters['%s'] = OutputParameter(%s)\n"\
                % (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

        for name, art in outputs.get("artifacts", {}).items():
            kwargs = self.get_kwargs(art, OutputArtifact)
            kwargs.pop("name", None)
            if "archive" in kwargs and kwargs["archive"] == config[
                    "archive_mode"]:
                del kwargs["archive"]
            _from = kwargs.get("_from")
            if isinstance(_from, str):
                match = input_artifact_pattern.match(_from)
                if match:
                    kwargs["_from"] = Variable(
                        "%s.inputs.artifacts['%s']" % (
                            var_name, match.group(1)))
                match = task_output_artifact_pattern.match(_from)
                if match:
                    kwargs["_from"] = Variable(
                        "%s.outputs.artifacts['%s']" % (
                            task_dict[match.group(1)], match.group(2)))
            self.imports.add(("dflow", "OutputArtifact"))
            self.code += "%s.outputs.artifacts['%s'] = OutputArtifact(%s)\n" %\
                (var_name, name, ", ".join(["%s=%s" % (k, evalable_repr(
                    v, self.imports)) for k, v in kwargs.items()]))

    def generate(self):
        self.graph = jsonpickle.loads(json.dumps(self.graph))
        entrypoint_type = self.graph["templates"][self.graph["entrypoint"]][
            "type"]

        self.templates = {}
        for name, template in list(self.graph["templates"].items()):
            if template["type"] == "PythonOPTemplate":
                var_name = self.get_var_name(name)
                self.render_python_op_template(var_name, template)
                self.templates[name] = var_name
                del self.graph["templates"][name]
            elif template["type"] == "ScriptOPTemplate":
                var_name = self.get_var_name(name)
                self.render_script_op_template(var_name, template)
                self.templates[name] = var_name
                del self.graph["templates"][name]

        while len(self.graph["templates"]) > 0:
            update = False
            for name, template in list(self.graph["templates"].items()):
                if template["type"] == "Steps":
                    if all([all([ps["template"] in self.templates or ps[
                        "template"] == name for ps in s])
                            for s in template["steps"]]):
                        var_name = self.get_var_name(name)
                        self.render_steps(var_name, template)
                        self.templates[name] = var_name
                        del self.graph["templates"][name]
                        update = True
                elif template["type"] == "DAG":
                    if all([t["template"] in self.templates or t[
                            "template"] == name for t in template["tasks"]]):
                        var_name = self.get_var_name(name)
                        self.render_dag(var_name, template)
                        self.templates[name] = var_name
                        del self.graph["templates"][name]
                        update = True
            assert update, "Failed to resolve templates: %s" % list(
                self.graph["templates"])

        del self.graph["templates"]
        entrypoint = self.templates[self.graph.pop("entrypoint")]
        kwargs = self.get_kwargs(self.graph, Workflow)
        if "namespace" in kwargs and kwargs["namespace"] == config[
                "namespace"]:
            del kwargs["namespace"]
        if entrypoint_type == "Steps":
            kwargs["steps"] = Variable(entrypoint)
        elif entrypoint_type == "DAG":
            kwargs["dag"] = Variable(entrypoint)
        self.imports.add(("dflow", "Workflow"))
        self.code += "wf = Workflow(%s)\n" % ", ".join(["%s=%s" % (
            k, evalable_repr(v, self.imports)) for k, v in kwargs.items()])

        module_imports = {}
        for m, i in self.imports:
            if m not in module_imports:
                module_imports[m] = []
            module_imports[m].append(i)
        import_code = ""
        for i in sorted(module_imports.pop(None, [])):
            import_code += "import %s\n" % i
        for m in sorted(module_imports):
            import_code += "from %s import %s\n" % (
                m, ", ".join(sorted(module_imports[m])))
        self.code = import_code + "\n" + self.code

        self.code += "wf.submit()\n"
        return self.code


def gen_code(graph):
    gen = CodeGenerator(graph)
    return gen.generate()
